package org.insa.graphs.algorithm.utils;

import java.util.ArrayList;

/**
 * Implements a binary heap containing elements of type E.
 *
 * Note that all comparisons are based on the compareTo method, hence E must
 * implement Comparable
 * 
 * @author Mark Allen Weiss
 * @author DLB
 */
public class BinaryHeap<E extends Comparable<E>> implements PriorityQueue<E> {

    // Number of elements in heap.
    private int currentSize;

    // The heap array.
    protected final ArrayList<E> array;

    /**
     * Construct a new empty binary heap.
     */
    public BinaryHeap() {
        this.currentSize = 0;
        this.array = new ArrayList<E>();
    }

    /**
     * Construct a copy of the given heap.
     * 
     * @param heap Binary heap to copy.
     */
    public BinaryHeap(BinaryHeap<E> heap) {
        this.currentSize = heap.currentSize;
        this.array = new ArrayList<E>(heap.array);
    }

    /**
     * Set an element at the given index.
     * 
     * @param index Index at which the element should be set.
     * @param value Element to set.
     */
    private void arraySet(int index, E value) {
        if (index == this.array.size()) {
            this.array.add(value);
        }
        else {
            this.array.set(index, value);
        }
    }

    /**
     * @return Index of the parent of the given index.
     */
    protected int indexParent(int index) {
        return (index - 1) / 2;
    }

    /**
     * @return Index of the left child of the given index.
     */
    protected int indexLeft(int index) {
        return index * 2 + 1;
    }

    /**
     * Internal method to percolate up in the heap.
     * 
     * @param index Index at which the percolate begins.
     */
    private void percolateUp(int index) {
        E x = this.array.get(index);

        for (; index > 0
                && x.compareTo(this.array.get(indexParent(index))) < 0; index = indexParent(
                        index)) {
            E moving_val = this.array.get(indexParent(index));
            this.arraySet(index, moving_val);
        }

        this.arraySet(index, x);
    }

    /**
     * Internal method to percolate down in the heap.
     * 
     * @param index Index at which the percolate begins.
     */
    private void percolateDown(int index) {
        int ileft = indexLeft(index);
        int iright = ileft + 1;

        if (ileft < this.currentSize) {
            E current = this.array.get(index);
            E left = this.array.get(ileft);
            boolean hasRight = iright < this.currentSize;
            E right = (hasRight) ? this.array.get(iright) : null;

            if (!hasRight || left.compareTo(right) < 0) {
                // Left is smaller
                if (left.compareTo(current) < 0) {
                    this.arraySet(index, left);
                    this.arraySet(ileft, current);
                    this.percolateDown(ileft);
                }
            }
            else {
                // Right is smaller
                if (right.compareTo(current) < 0) {
                    this.arraySet(index, right);
                    this.arraySet(iright, current);
                    this.percolateDown(iright);
                }
            }
        }
    }

    @Override
    public boolean isEmpty() {
        return this.currentSize == 0;
    }

    @Override
    public int size() {
        return this.currentSize;
    }
    
    public boolean isValid() {

    	if(this.isEmpty()) {
    		return true;
    	}
    	
    	return isValid(0);
    }
    
    private boolean isValid(int indexStart) {
    	//on doit vérifier que l'index est plus petit que ses deux parents
    	
        int ileft = indexLeft(indexStart);
        int iright = ileft + 1;
        boolean hasRight = iright < this.currentSize;
        boolean hasLeft = ileft < this.currentSize;
       
        
        if(hasRight) {
        	//si on a un point à droite on vérifie que le point est valide, puis que tous les points en dessous à droite le sont.
        	if(this.array.get(indexStart).compareTo(this.array.get(iright)) > 0)return false;
        	if(!isValid(iright))return false;
        }
        if(hasLeft) {
        	//si on a un point à gauche on vérifie que le point est valide, puis que tous les points en dessous à gauche le sont aussi.
        	if(this.array.get(indexStart).compareTo(this.array.get(ileft)) > 0)return false;
        	if(!isValid(ileft))return false;
        }

        //ici, on a vérifié une branche ou les deux ou zéro, et on a pas fail donc c'est tout bon !
        return true;
    	
    	
    }

    @Override
    public void insert(E x) {
        int index = this.currentSize++;
        this.arraySet(index, x);
        this.percolateUp(index);
    }
    
    // recursive version
    private int find(int indexStart, E x) throws ElementNotFoundException {
    	
    	//it's a heap so uses it to find a log(n) speed the element	
    	
        int ileft = indexLeft(indexStart);
        int iright = ileft + 1;
        boolean hasRight = iright < this.currentSize;
        boolean hasLeft = ileft < this.currentSize;
        
        if(hasRight) {
        	try {
        		if(x.compareTo(this.array.get(iright)) == 0) {
        			return iright;
        		}
        		else if(x.compareTo(this.array.get(iright)) > 0) {
            		return find(iright, x);
            	}
        	}catch (ElementNotFoundException e) {
        		//do nothing...
        	}
        }
        
        if(hasLeft) {
        	try {
        		if(x.compareTo(this.array.get(ileft)) == 0) {
        			return ileft;
        		}
        		else if(x.compareTo(this.array.get(ileft)) > 0) {
            		return find(ileft, x);
            	}
        	}catch (ElementNotFoundException e) {
        		//do nothing...
        	}
        }
        
        //here we dont find anything from this index, throw exception
        throw new ElementNotFoundException(x);
    }
    
    @Override
    public int find(E x) throws ElementNotFoundException {
    	
    	if(this.currentSize == 0) {
    		throw new ElementNotFoundException(x);
    	}
    	
    	
    	/*int index =  this.array.indexOf(x);
    	
    	if(index < 0 || index >= this.currentSize) {
    		throw new ElementNotFoundException(x);
    	}
    	else {
    		return index;
    	}*/
    	
    	//try to see if it's root
    	if(this.array.get(0).compareTo(x) == 0) {
    		return 0;
    	}
    	
    	//so now we need to find deeper.
		return find(0, x);
    }

    @Override
    public void remove(E x) throws ElementNotFoundException {
        // first step, find the element on the heap
    	
    	int myIndex = find(x);
    	
    	
    	//now we have the element of the object to remove
    	//so it's like deleting root, but instead of root it's index of x
    	
    	E lastItem = this.array.get(--this.currentSize);
        this.arraySet(myIndex, lastItem);
        this.percolateUp(myIndex);
        this.percolateDown(myIndex);
    	
    	
    }

    @Override
    public E findMin() throws EmptyPriorityQueueException {
        if (isEmpty())
            throw new EmptyPriorityQueueException();
        return this.array.get(0);
    }

    @Override
    public E deleteMin() throws EmptyPriorityQueueException {
        E minItem = findMin();
        E lastItem = this.array.get(--this.currentSize);
        this.arraySet(0, lastItem);
        this.percolateDown(0);
        return minItem;
    }

    /**
     * Creates a multi-lines string representing a sorted view of this binary heap.
     * 
     * @return a string containing a sorted view this binary heap.
     */
    public String toStringSorted() {
        return BinaryHeapFormatter.toStringSorted(this, -1);
    }

    /**
     * Creates a multi-lines string representing a sorted view of this binary heap.
     * 
     * @param maxElement Maximum number of elements to display. or {@code -1} to
     *                   display all the elements.
     * 
     * @return a string containing a sorted view this binary heap.
     */
    public String toStringSorted(int maxElement) {
        return BinaryHeapFormatter.toStringSorted(this, maxElement);
    }

    /**
     * Creates a multi-lines string representing a tree view of this binary heap.
     * 
     * @return a string containing a tree view of this binary heap.
     */
    public String toStringTree() {
        return BinaryHeapFormatter.toStringTree(this, Integer.MAX_VALUE);
    }

    /**
     * Creates a multi-lines string representing a tree view of this binary heap.
     * 
     * @param maxDepth Maximum depth of the tree to display.
     * 
     * @return a string containing a tree view of this binary heap.
     */
    public String toStringTree(int maxDepth) {
        return BinaryHeapFormatter.toStringTree(this, maxDepth);
    }

    @Override
    public String toString() {
        return BinaryHeapFormatter.toStringTree(this, 8);
    }

}
